var dom = require("dom-tree");
var on = require("dom-event");
on.child = require("component-delegate");

var select = require("dom-select");
var style = require("dom-style");
var classes = require("dom-classes");
var format = require("format-text");
var bindKey = require("key-event");
var escape = require("escape-html");
var failingCode = require("failing-code");
var socket = require("./socket");
var grep = require("./grep");
var templates = require("./templates");
var options = require("./options");
var url;

on(window, 'resize', updatePositions);
on(window, 'hashchange', run);
on(window, 'scroll', saveScrollState);
socket(updateConn);

module.exports = {
  addError: addError,
  markTest: markTest,
  pass: pass,
  status: status,
  run: run,
  list: list
};

function addError (error) {
  var top = select('.top');
  var view = {
    name: error.name
  };

  if (!top) {
    setup();
    top = select('.top');
  };

  classes.add(select('.top'), 'failed');
  view.stack = format(templates.stack, escape(error.stack).replace(/\n\s+/g, templates['stack-line']));

  view.code = error.source.length ? formatCode(error.source) : '';

  if (error.expected != undefined) {
    view.diff = format(templates.diff, escape(JSON.stringify(error.expected, null, " ")), escape(JSON.stringify(error.actual, null, " "))) || " ";
    view['diff-class'] = '';
  } else {
    view.diff = "";
    view['diff-class'] = ' empty';
  }

  if (addError.last != error.test) {
    view.title = '<h3>' + error.test +'</h3>';
    addError.last = error.test;
  } else {
    view.title = '';
  }

  if (!addError.counter) {
    addError.counter = 1;
  }

  view.id = addError.counter++;

  dom.add(select('.results .errors'), templates.error, view);
  var el = select('#error-' + view.id);
  on.child.bind(el, '.stack-line', 'click', function (e) {
    var index = Array.prototype.indexOf.call(e.target.parentElement.children, e.target) - 1;
    select('.code', el).innerHTML = formatCode(failingCode(error, __source_code, index));
  });
}

function markTest (test) {
  var overview = select('.overview');
  if (overview) return;

  var list = select('.waiting .list');

  if (!list) {
    dom.add(select('.waiting h1'), templates.overview);
    list = select('.waiting .list');
  }

  dom.add(list, templates.test, {
    icon: '',
    name: test.name
  });
}

function setup () {
  dom.remove(select('.waiting'));

  var template = format(templates['layout'], templates);
  dom.add(document.body, template, {
    grep: grep() || '',
    conn: ''
  });

  on(select('.frame-button'), 'click', toggleFrame);
  on(select('.run-again'), 'click', run);
  on(select('.maximize'), 'click', maximize);
  on(select('.minimize'), 'click', minimize);

  updateConn();

  setupGrep();

  if (localStorage['frame-open']) {
    toggleFrame();
  }

  if (localStorage['frame-maximized']) {
    maximize();
  }

  updateFramePosition();
  recoverScrollPosition();
}

function setupGrep () {
  var el = select('#grep');

  on(select('.grep label'), 'click', function () {
    el.focus();
  });

  bindKey(el, 'enter', function () {
    grep(el.value);
  });
}

function run (_url) {
  if (typeof _url == 'string') {
    url = _url;
  }

  addError.last = undefined;
  status('running');

  dom.add(document.body, templates['frame'], {
    url: url,
    options: options.stringify()
  });
}

function end () {
  classes.remove(select('.waiting'), '.running');
}

function pass (assertions) {
  setup();
  classes.add(select('.top'), 'passed');
  select('.pass').innerHTML = format(templates.pass, assertions);
  classes.add(select('.results'), 'passed');
}

function toggleFrame () {
  var isOpen = classes.has(select('.results'), 'open');

  if (isOpen) {
    closeFrame();
  } else {
    openFrame();
  }

  updateFramePosition();
}

function openFrame () {
  var results = select('.results');
  var frame = select('.frame');

  classes.add(results, 'open');
  classes.add(frame, 'open');

  localStorage['frame-open'] = true;
}

function closeFrame () {
  var results = select('.results');
  var frame = select('.frame');

  classes.remove(results, 'open');
  classes.remove(frame, 'open');

  delete localStorage['frame-open'];
}

function updatePositions () {
  updateFramePosition();
}

function updateFramePosition () {
  var frame = select('.frame');
  var results = select('.results');

  if (!frame) return;

  var isOpen = classes.has(frame, 'open') && !classes.has(select('body'), 'maximized');
  var right = isOpen ? results.offsetWidth : 0;

  style(select('.frame-button'), 'right', right + 'px');

  if (!isOpen) return;

  style(frame, {
    width: results.offsetWidth - 1 + 'px',
    left: right + 'px',
    height: '100%'
  });
}

function status (msg) {
  var el = select('.status');

  if (!el) {
    document.body.innerHTML = format(templates.waiting, {
      message: msg
    });
    return;
  }

  el.innerHTML = msg;
  select('.waiting').className = 'waiting center ' + msg;
}

function list (tests) {
  var key;
  for (key in tests) {
    dom.add(select('.overview .list'), templates.test, {
      icon: tests[key] ?  '✓' : '✖',
      name: key
    });
  }
}

function updateConn (msg) {
  var el = select('.conn');
  if (!el) return;

  el.innerHTML = socket.isOpen() ? 'Watching File Changes' : 'Disconnected';
}

function maximize () {
  classes.add(select('body'), 'maximized');
  localStorage['frame-maximized'] = true;
  updateFramePosition();
}

function minimize () {
  delete localStorage['frame-maximized'];
  classes.remove(select('body'), 'maximized');
  updateFramePosition();
}

function saveScrollState () {
  if (saveScrollState.defer) {
    clearTimeout(saveScrollState.defer);
    saveScrollState.defer = undefined;
  }

  saveScrollState.defer = setTimeout(function () {
    localStorage['scrollTop'] = document.body.scrollTop;
  }, 500);
}

function recoverScrollPosition () {
  document.body.scrollTop = Number(localStorage['scrollTop']);
}

function formatCode (source) {
  return format(templates.code, {
    'first-line-num': source[0].line,
    'first-line-source': escape(source[0].code),
    'second-line-num': source[1].line,
    'second-line-source': escape(source[1].code),
    'third-line-num': source[2].line,
    'third-line-source': escape(source[2].code)
  })
}
